"""SKALE chain test"""

import pytest
from eth_typing import HexStr
from hexbytes import HexBytes
from web3 import Web3

from skale.contracts.manager.schains import SchainStructure
from skale.dataclasses.schain_options import AllocationType, SchainOptions
from skale.skale_manager import SkaleManager
from skale.utils.contracts_provision.fake_multisig_contract import (
    FAKE_MULTISIG_DATA_PATH,
)
from skale.utils.contracts_provision.main import (
    create_schain,
    generate_random_schain_data,
)
from skale.utils.helper import get_abi, schain_name_to_hash
from skale.wallets.web3_wallet import generate_wallet
from tests.constants import (
    DEFAULT_NODE_NAME,
    DEFAULT_SCHAIN_HASH,
    DEFAULT_SCHAIN_NAME,
    LIFETIME_SECONDS,
    SCHAIN_FIELDS,
)


def test_get(skale):
    schain = skale.schains.get(DEFAULT_SCHAIN_HASH)
    assert isinstance(schain, SchainStructure)
    assert isinstance(schain.options, SchainOptions)


def test_get_by_name(skale, schain):
    schain_by_name = skale.schains.get_by_name(schain)
    assert isinstance(schain_by_name, SchainStructure)
    assert schain_by_name.name == schain
    assert schain_by_name.index_in_owner_list == 0
    assert schain_by_name.part_of_node == 1
    assert schain_by_name.lifetime == 3600
    assert schain_by_name.deposit == 0
    assert schain_by_name.generation == 0
    assert schain_by_name.options == SchainOptions(
        multitransaction_mode=False,
        threshold_encryption=False,
        allocation_type=AllocationType.DEFAULT,
        external_gas_difficulty=HexStr('0x01'),
        min_gas_price=None,
        max_gas_price=None,
    )


def test_schain_get_plain(skale):
    schain = skale.schains.get(DEFAULT_SCHAIN_HASH)
    for field in SCHAIN_FIELDS:
        assert hasattr(schain, field), f'Missing field: {field}'


def test_schain_get_object(skale, schain):
    schain_hash = schain_name_to_hash(schain)
    schain_struct = skale.schains.get(schain_hash)
    assert isinstance(schain_struct, SchainStructure)
    assert isinstance(schain_struct.options, SchainOptions)
    assert schain_struct.name == schain
    assert schain_struct.index_in_owner_list == 0
    assert schain_struct.part_of_node == 1
    assert schain_struct.lifetime == 3600
    assert schain_struct.deposit == 0
    assert schain_struct.generation == 0
    assert schain_struct.options == SchainOptions(
        multitransaction_mode=False,
        threshold_encryption=False,
        allocation_type=AllocationType.DEFAULT,
        external_gas_difficulty=HexStr('0x01'),
        min_gas_price=None,
        max_gas_price=None,
    )


def test_schains_for_owner(skale, schain, empty_account):
    schains = skale.schains.schains_for_owner(skale.wallet.address)
    assert isinstance(schains, list)
    assert schains[-1].mainnet_owner == skale.wallet.address

    schains = skale.schains.schains_for_owner(empty_account.address)
    assert schains == []


def test_schains_for_node(skale, schain):
    node_id = skale.nodes.node_name_to_index(DEFAULT_NODE_NAME)
    schains_for_node = skale.schains.schains_for_node(node_id)
    schain_hashes_for_node = skale.schains_internal.schain_hashes_for_node(node_id)

    assert isinstance(schains_for_node, list)
    assert len(schains_for_node) > 0
    assert len(schains_for_node) == len(schain_hashes_for_node)
    test_schain = schains_for_node[0]
    schain_node_ids = skale.schains_internal.node_ids_for_schain(test_schain.name)

    assert node_id in schain_node_ids


def test_schain_name_to_hash():
    schain_hash = schain_name_to_hash(DEFAULT_SCHAIN_NAME)
    assert schain_hash == Web3.to_bytes(hexstr=DEFAULT_SCHAIN_HASH)


def test_all_schain_hashes(skale, schain):
    schains_ids = skale.schains_internal.all_schain_hashes()
    schain_struct = skale.schains.get(schains_ids[-1])
    assert schain_struct.name == schain
    assert schain_struct.index_in_owner_list == 0
    assert schain_struct.part_of_node == 1
    assert schain_struct.lifetime == 3600
    assert schain_struct.deposit == 0
    assert schain_struct.generation == 0
    assert schain_struct.options == SchainOptions(
        multitransaction_mode=False,
        threshold_encryption=False,
        allocation_type=AllocationType.DEFAULT,
        external_gas_difficulty=HexStr('0x01'),
        min_gas_price=None,
        max_gas_price=None,
    )


def test_schain_price(skale):
    schain_price = skale.schains.schain_price(1, LIFETIME_SECONDS)
    assert schain_price > 0
    assert isinstance(schain_price, int)


def test_add_schain_by_foundation(skale, nodes):
    skale.schains.grant_role(skale.schains.schain_creator_role(), skale.wallet.address)
    _, lifetime_seconds, name = generate_random_schain_data(skale)
    type_of_nodes = 1  # test2 schain
    try:
        skale.schains.add_schain_by_foundation(lifetime_seconds, type_of_nodes, 0, name)
        schains_ids_after = skale.schains_internal.all_schain_hashes()
        schains_names = [skale.schains.get(sid).name for sid in schains_ids_after]
        assert name in schains_names
        new_schain = skale.schains.get_by_name(name)
        assert new_schain.mainnet_owner == skale.wallet.address

        schain = skale.schains.get_by_name(name)
        assert schain.options.multitransaction_mode is False
        assert schain.options.threshold_encryption is False
        assert schain.options.allocation_type is AllocationType.DEFAULT
        assert schain.options.external_gas_difficulty == HexStr('0x01')
        assert schain.options.min_gas_price is None
        assert schain.options.max_gas_price is None
    finally:
        skale.manager.delete_schain(name, wait_for=True)

    schains_ids_after = skale.schains_internal.all_schain_hashes()

    schains_names = [skale.schains.get(sid).name for sid in schains_ids_after]
    assert name not in schains_names


@pytest.mark.parametrize(
    'mts,threshold,alloc,pow_difficulty,min_gas,max_gas',
    [
        (True, False, AllocationType.MAX_CONSENSUS_DB, HexStr('0x01'), None, None),
        (
            False,
            True,
            AllocationType.MAX_FILESTORAGE,
            HexStr('0x02'),
            256,
            512,
        ),
        (False, False, AllocationType.DEFAULT, HexStr('0x01'), 4096, None),
    ],
)
def test_add_schain_by_foundation_with_options(
    mts, threshold, alloc, pow_difficulty, min_gas, max_gas, skale, nodes
):
    skale.schains.grant_role(skale.schains.schain_creator_role(), skale.wallet.address)
    _, lifetime_seconds, name = generate_random_schain_data(skale)
    type_of_nodes = 1  # test2 schain
    try:
        skale.schains.add_schain_by_foundation(
            lifetime_seconds,
            type_of_nodes,
            0,
            name,
            options=SchainOptions(
                multitransaction_mode=mts,
                threshold_encryption=threshold,
                allocation_type=alloc,
                external_gas_difficulty=pow_difficulty,
                min_gas_price=min_gas,
                max_gas_price=max_gas,
            ),
            wait_for=True,
        )
        schain = skale.schains.get_by_name(name)

        assert schain.options.multitransaction_mode is mts
        assert schain.options.threshold_encryption is threshold
        assert schain.options.allocation_type is alloc
        assert schain.options.external_gas_difficulty == pow_difficulty
        assert schain.options.min_gas_price == min_gas
        assert schain.options.max_gas_price == max_gas
    finally:
        skale.manager.delete_schain(name)


def test_add_schain_by_foundation_custom_owner(skale, nodes):
    skale.schains.grant_role(skale.schains.schain_creator_role(), skale.wallet.address)
    _, lifetime_seconds, name = generate_random_schain_data(skale)
    type_of_nodes = 1  # test2 schain
    main_wallet = skale.wallet
    custom_wallet = generate_wallet(skale.web3)
    try:
        skale.schains.add_schain_by_foundation(
            lifetime_seconds,
            type_of_nodes,
            0,
            name,
            schain_owner=custom_wallet.address,
            wait_for=True,
        )

        new_schain = skale.schains.get_by_name(name)
        assert new_schain.mainnet_owner != skale.wallet.address
        assert new_schain.mainnet_owner == custom_wallet.address

        skale.wallet = custom_wallet
    finally:
        skale.wallet = main_wallet
        skale.manager.delete_schain_by_root(name, wait_for=True)

    schains_ids_after = skale.schains_internal.all_schain_hashes()

    schains_names = [skale.schains.get(sid).name for sid in schains_ids_after]
    assert name not in schains_names


def test_add_schain_by_foundation_custom_originator(skale: SkaleManager, nodes):
    skale.schains.grant_role(skale.schains.schain_creator_role(), skale.wallet.address)
    _, lifetime_seconds, name = generate_random_schain_data(skale)
    type_of_nodes = 1  # test2 schain
    custom_originator = generate_wallet(skale.web3)

    fake_multisig_data = get_abi(FAKE_MULTISIG_DATA_PATH)
    payable_contract_address = fake_multisig_data['address']

    try:
        skale.schains.add_schain_by_foundation(
            lifetime_seconds,
            type_of_nodes,
            0,
            name,
            schain_owner=payable_contract_address,
            schain_originator=custom_originator.address,
        )
        new_schain = skale.schains.get_by_name(name)

        assert new_schain.originator != skale.wallet.address
        assert new_schain.originator == custom_originator.address

    finally:
        if name:
            skale.manager.delete_schain_by_root(name)

        schains_ids_after = skale.schains_internal.all_schain_hashes()

    schains_names = [skale.schains.get(sid).name for sid in schains_ids_after]
    assert name not in schains_names


def test_active_schains_for_node(skale, nodes, schain):
    name = None
    try:
        name = create_schain(skale, random_name=True)
        node_id = skale.nodes.node_name_to_index(DEFAULT_NODE_NAME)
        active_schains = skale.schains.active_schains_for_node(node_id)
        all_schains = skale.schains.schains_for_node(node_id)
        all_active_schains = [schain for schain in all_schains if schain.active]
        for active_schain in all_active_schains:
            assert active_schain in active_schains
    finally:
        if name:
            skale.manager.delete_schain_by_root(name)


def test_name_to_group_id(skale):
    name = 'TEST'
    gid = skale.schains.name_to_group_id(name)
    assert gid == HexBytes('0x852daa74cc3c31fe64542bb9b8764cfb91cc30f9acf9389071ffb44a9eefde46')  # noqa


def test_options(skale, nodes):
    schain_options = SchainOptions(
        multitransaction_mode=True,
        threshold_encryption=False,
        allocation_type=AllocationType.DEFAULT,
        external_gas_difficulty=HexStr('0x01'),
        min_gas_price=256,
        max_gas_price=512,
    )
    name = None
    try:
        name = create_schain(skale, random_name=True, schain_options=schain_options)
        schain_hash = schain_name_to_hash(name)
        options = skale.schains.options(schain_hash)
        assert options == schain_options
        options = skale.schains.options_by_name(name)
        assert options == schain_options
        raw_options = skale.schains._SChains__raw_options(schain_hash)
        assert raw_options == [
            ('multitr', b'\x01'),
            ('encrypt', b'\x00'),
            ('alloc', b'\x00'),
            ('powdifficulty', b'\x01'),
            ('mingasprice', b'\x01\x00'),
            ('maxgasprice', b'\x02\x00'),
        ]

    finally:
        if name:
            skale.manager.delete_schain(name)


def test_restart_schain_creation(skale, nodes, schain):
    name = schain
    skale.schains.restart_schain_creation(name)



"""SKALE test multithreading"""

import threading

from skale import SkaleManager
from skale.utils.helper import get_skale_manager_address
from skale.utils.web3_utils import init_web3
from tests.constants import DEFAULT_NODE_NAME, ENDPOINT, TEST_ABI_FILEPATH


def get_node_data():
    skale = SkaleManager(ENDPOINT, get_skale_manager_address(TEST_ABI_FILEPATH))
    for _ in range(0, 30):
        skale.nodes.get_by_name(DEFAULT_NODE_NAME)


def test_multithread_calls():
    init_web3(ENDPOINT)
    monitors = []
    for _ in range(0, 5):
        monitor = threading.Thread(target=get_node_data, daemon=True)
        monitor.start()
        monitors.append(monitor)
    for monitor in monitors:
        monitor.join()



from unittest import mock

import pytest
from web3 import Web3

from skale import SkaleManager
from skale.transactions.exceptions import (
    DryRunFailedError,
    TransactionFailedError,
    TransactionNotMinedError,
)
from skale.transactions.tools import (
    TxCallResult,
    TxStatus,
    estimate_gas,
    get_block_gas_limit,
    run_tx_with_retry,
)
from skale.utils.account_tools import generate_account
from skale.utils.helper import get_skale_manager_address
from skale.utils.web3_utils import init_web3, wait_for_receipt
from skale.wallets import Web3Wallet
from skale.wallets.web3_wallet import generate_wallet
from tests.constants import (
    D_VALIDATOR_DESC,
    D_VALIDATOR_FEE,
    D_VALIDATOR_MIN_DEL,
    D_VALIDATOR_NAME,
    ENDPOINT,
    TEST_ABI_FILEPATH,
)

ETH_IN_WEI = 10**18
RETRIES_NUMBER = 2


def generate_new_skale():
    web3 = init_web3(ENDPOINT)
    account = generate_account(web3)
    wallet = Web3Wallet(account['private_key'], web3)
    wallet.wait = mock.Mock()
    return SkaleManager(ENDPOINT, get_skale_manager_address(TEST_ABI_FILEPATH), wallet)


def test_run_tx_with_retry(skale):
    account = generate_account(skale.web3)
    address_to = account['address']
    address_from = Web3.to_checksum_address(skale.wallet.address)
    address_to = Web3.to_checksum_address(address_to)
    balance_from_before = skale.token.get_balance(address_from)
    balance_to_before = skale.token.get_balance(address_to)

    token_amount = 10 * ETH_IN_WEI
    tx_res = run_tx_with_retry(
        skale.token.transfer, account['address'], token_amount, wait_for=True, max_retries=5
    )
    tx_res.raise_for_status()

    balance_from_after = skale.token.get_balance(address_from)
    assert balance_from_after == balance_from_before - token_amount
    balance_to_after = skale.token.get_balance(address_to)
    assert balance_to_after == balance_to_before + token_amount


def test_run_tx_with_retry_dry_run_failed(skale):
    dry_run_call_mock = mock.Mock(
        return_value=TxCallResult(
            status=TxStatus.FAILED, error='revert', message='Dry run test failure', data={}
        )
    )
    account = generate_account(skale.web3)
    token_amount = 10 * ETH_IN_WEI
    with mock.patch('skale.contracts.base_contract.make_dry_run_call', dry_run_call_mock):
        tx_res = run_tx_with_retry(
            skale.token.transfer,
            account['address'],
            token_amount,
            wait_for=True,
            raise_for_status=False,
            max_retries=RETRIES_NUMBER,
        )
        with pytest.raises(DryRunFailedError):
            tx_res.raise_for_status()

    assert dry_run_call_mock.call_count == RETRIES_NUMBER


def test_run_tx_with_retry_tx_failed(failed_skale):
    skale = failed_skale
    account = generate_account(skale.web3)
    token_amount = 10 * ETH_IN_WEI
    tx_res = run_tx_with_retry(
        skale.token.transfer,
        account['address'],
        token_amount,
        wait_for=True,
        raise_for_status=False,
        max_retries=RETRIES_NUMBER,
    )
    with pytest.raises(TransactionFailedError):
        tx_res.raise_for_status()

    assert skale.wallet.sign_and_send.call_count == RETRIES_NUMBER
    assert skale.wallet.wait.call_count == RETRIES_NUMBER


def test_run_tx_with_retry_insufficient_balance(skale):
    sender_skale = generate_new_skale()
    token_amount = 10 * ETH_IN_WEI
    tx_res = run_tx_with_retry(
        sender_skale.token.transfer,
        skale.wallet.address,
        token_amount,
        raise_for_status=False,
        max_retries=RETRIES_NUMBER,
    )

    assert tx_res.attempts == RETRIES_NUMBER


def test_estimate_gas(skale):
    main_wallet = skale.wallet
    skale.wallet = generate_wallet(skale.web3)

    method = skale.validator_service.contract.functions.registerValidator(
        D_VALIDATOR_NAME, D_VALIDATOR_DESC, D_VALIDATOR_FEE, D_VALIDATOR_MIN_DEL
    )
    opts = {'from': skale.wallet.address, 'value': 0}

    block_gas_limit = get_block_gas_limit(skale.web3)
    estimated_gas = estimate_gas(skale.web3, method, opts)

    assert isinstance(estimated_gas, int)
    assert estimated_gas != block_gas_limit

    with mock.patch.object(method, 'estimate_gas', return_value=10000000000):
        estimated_gas = estimate_gas(skale.web3, method, opts)

    assert estimated_gas == block_gas_limit
    skale.wallet = main_wallet


@pytest.mark.skip(reason='This test is not working on the CI')
def test_tx_fee_options(skale):
    account = generate_account(skale.web3)
    address_to = account['address']
    address_from = Web3.to_checksum_address(skale.wallet.address)
    address_to = Web3.to_checksum_address(address_to)
    balance_from_before = skale.token.get_balance(address_from)
    balance_to_before = skale.token.get_balance(address_to)
    token_amount = 10 * ETH_IN_WEI

    max_fee = 10**9
    max_priority_fee = 10**9
    res = skale.token.transfer(
        account['address'],
        token_amount,
        max_fee_per_gas=max_fee,
        max_priority_fee_per_gas=max_priority_fee,
    )
    r = res.receipt
    assert r['effectiveGasPrice'] == max_fee

    balance_from_after = skale.token.get_balance(address_from)
    assert balance_from_after == balance_from_before - token_amount
    balance_to_after = skale.token.get_balance(address_to)
    assert balance_to_after == balance_to_before + token_amount

    skale.token.transfer(
        account['address'],
        token_amount,
        max_fee_per_gas=max_fee,
        max_priority_fee_per_gas=int(max_fee / 2),
    )


def test_wait_for_receipt_fail(skale):
    with pytest.raises(TransactionNotMinedError):
        wait_for_receipt(skale.web3, b'\x00' * 32, timeout=2)



"""Tests for skale/allocator/escrow.py"""

from skale.types.delegation import DelegationStatus
from skale.utils.account_tools import check_skale_balance, send_eth
from skale.utils.contracts_provision import D_PLAN_ID, MONTH_IN_SECONDS
from skale.utils.contracts_provision.allocator import connect_test_beneficiary
from skale.utils.contracts_provision.main import _skip_evm_time
from skale.wallets.web3_wallet import generate_wallet
from tests.constants import (
    D_DELEGATION_AMOUNT,
    D_DELEGATION_INFO,
    D_DELEGATION_PERIOD,
    D_VALIDATOR_ID,
)
from tests.manager.delegation.delegation_controller_test import _get_number_of_delegations


def _delegate_via_escrow(skale_allocator, wallet):
    main_wallet = skale_allocator.wallet
    send_eth(skale_allocator.web3, main_wallet, wallet.address, 0.5)
    connect_test_beneficiary(skale_allocator, D_PLAN_ID, wallet)

    skale_allocator.allocator.start_vesting(wallet.address, wait_for=True)
    skale_allocator.wallet = wallet

    skale_allocator.escrow.delegate(
        validator_id=D_VALIDATOR_ID,
        amount=D_DELEGATION_AMOUNT,
        delegation_period=D_DELEGATION_PERIOD,
        info=D_DELEGATION_INFO,
        beneficiary_address=skale_allocator.wallet.address,
        wait_for=True,
    )
    skale_allocator.wallet = main_wallet


def test_delegate(skale, skale_allocator):
    main_wallet = skale_allocator.wallet
    wallet = generate_wallet(skale_allocator.web3)
    send_eth(skale_allocator.web3, main_wallet, wallet.address, 0.1)

    connect_test_beneficiary(skale_allocator, D_PLAN_ID, wallet)
    skale_allocator.allocator.start_vesting(wallet.address, wait_for=True)

    _skip_evm_time(skale.web3, MONTH_IN_SECONDS * 12)

    skale_allocator.wallet = wallet

    num_of_delegations_before = _get_number_of_delegations(skale)
    skale_allocator.escrow.delegate(
        validator_id=D_VALIDATOR_ID,
        amount=D_DELEGATION_AMOUNT,
        delegation_period=D_DELEGATION_PERIOD,
        info=D_DELEGATION_INFO,
        beneficiary_address=skale_allocator.wallet.address,
        wait_for=True,
    )
    num_of_delegations_after = _get_number_of_delegations(skale)
    assert num_of_delegations_after == num_of_delegations_before + 1

    skale_allocator.wallet = main_wallet


def test_request_undelegate(skale, skale_allocator):
    main_wallet = skale_allocator.wallet
    wallet = generate_wallet(skale_allocator.web3)

    _delegate_via_escrow(skale_allocator, wallet)

    delegations = skale.delegation_controller.get_all_delegations_by_validator(
        validator_id=D_VALIDATOR_ID
    )
    delegation_id = delegations[-1]['id']
    skale.delegation_controller.accept_pending_delegation(delegation_id, wait_for=True)

    _skip_evm_time(skale.web3, MONTH_IN_SECONDS * (D_DELEGATION_PERIOD + 1))

    skale_allocator.wallet = wallet
    skale_allocator.escrow.request_undelegation(
        delegation_id, beneficiary_address=skale_allocator.wallet.address, wait_for=True
    )
    skale_allocator.wallet = main_wallet

    delegations = skale.delegation_controller.get_all_delegations_by_validator(
        validator_id=D_VALIDATOR_ID
    )
    assert delegations[-1]['id'] == delegation_id
    assert delegations[-1]['status'] == DelegationStatus.UNDELEGATION_REQUESTED


def test_retrieve(skale, skale_allocator):
    main_wallet = skale_allocator.wallet
    wallet = generate_wallet(skale_allocator.web3)

    _delegate_via_escrow(skale_allocator, wallet)

    delegations = skale.delegation_controller.get_all_delegations_by_validator(
        validator_id=D_VALIDATOR_ID
    )
    delegation_id = delegations[-1]['id']
    skale.delegation_controller.accept_pending_delegation(delegation_id, wait_for=True)

    _skip_evm_time(skale.web3, MONTH_IN_SECONDS * (D_DELEGATION_PERIOD + 1))

    skale_allocator.wallet = wallet
    skale_allocator.escrow.retrieve(
        beneficiary_address=skale_allocator.wallet.address, wait_for=True
    )
    skale_allocator.wallet = main_wallet
    # todo: improve test


def test_withdraw_bounty(skale, skale_allocator):
    main_wallet = skale_allocator.wallet
    wallet = generate_wallet(skale_allocator.web3)

    _delegate_via_escrow(skale_allocator, wallet)

    delegations = skale.delegation_controller.get_all_delegations_by_validator(
        validator_id=D_VALIDATOR_ID
    )
    delegation_id = delegations[-1]['id']
    skale.delegation_controller.accept_pending_delegation(delegation_id, wait_for=True)

    _skip_evm_time(skale.web3, MONTH_IN_SECONDS * (D_DELEGATION_PERIOD + 1))

    skale_allocator.wallet = wallet

    check_skale_balance(skale, wallet.address)
    skale_allocator.escrow.withdraw_bounty(
        D_VALIDATOR_ID,
        wallet.address,
        beneficiary_address=skale_allocator.wallet.address,
        wait_for=True,
    )
    skale_allocator.wallet = main_wallet
    # todo: improve test


def test_cancel_pending_delegation(skale_allocator, skale):
    main_wallet = skale_allocator.wallet
    wallet = generate_wallet(skale_allocator.web3)

    _delegate_via_escrow(skale_allocator, wallet)

    delegations = skale.delegation_controller.get_all_delegations_by_validator(
        validator_id=D_VALIDATOR_ID
    )
    delegation_id = delegations[-1]['id']
    skale_allocator.wallet = wallet

    skale_allocator.escrow.cancel_pending_delegation(
        delegation_id=delegation_id,
        beneficiary_address=skale_allocator.wallet.address,
        wait_for=True,
    )
    skale_allocator.wallet = main_wallet

    delegations = skale.delegation_controller.get_all_delegations_by_validator(
        validator_id=D_VALIDATOR_ID
    )
    assert delegations[-1]['id'] == delegation_id
    assert delegations[-1]['status'] == DelegationStatus.CANCELED



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

import logging
from typing import TYPE_CHECKING, Generic, TypeVar

from skale_contracts.types import ContractName

from skale.contracts.base_contract import BaseContract
from skale.skale_base import SkaleBase
from skale.wallets import BaseWallet

if TYPE_CHECKING:
    pass


logger = logging.getLogger(__name__)


SkaleType = TypeVar('SkaleType', bound=SkaleBase)


class SkaleContract(BaseContract, Generic[SkaleType]):
    def __init__(self, skale: SkaleType, name: ContractName):
        self.skale = skale
        self.name = name
        self.web3 = skale.web3
        self.init_contract(contract_name=name)

    def init_contract(self, contract_name: ContractName) -> None:
        address = self.skale.instance.get_contract_address(contract_name)
        abi = self.skale.instance.abi[contract_name]
        self._init_contract(address, abi)

    @property
    def wallet(self) -> BaseWallet:
        return self.skale.wallet

    @wallet.setter
    def wallet(self, value: BaseWallet) -> None:
        self.skale.wallet = value



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.


from eth_typing import ChecksumAddress
from web3.contract.contract import ContractFunction
from web3.types import TxParams

from skale.contracts.base_contract import transaction_method
from skale.contracts.skale_contract import SkaleContract
from skale.types.schain import SchainName
from skale.utils.helper import schain_name_to_hash


class CommunityPool(SkaleContract):
    @transaction_method
    def recharge_user_wallet(
        self, schain_name: SchainName, address: ChecksumAddress
    ) -> ContractFunction:
        return self.contract.functions.rechargeUserWallet(schain_name, address)

    @transaction_method
    def withdraw_funds(self, schain_name: SchainName, amount: int) -> ContractFunction:
        return self.contract.functions.withdrawFunds(schain_name, amount)

    @transaction_method
    def set_min_transaction_gas(self, min_gas_value: int) -> ContractFunction:
        return self.contract.functions.setMinTransactionGas(min_gas_value)

    @transaction_method
    def set_multiplier(self, new_numerator: int, new_divider: int) -> ContractFunction:
        return self.contract.functions.setMultiplier(new_numerator, new_divider)

    def get_balance(self, address: ChecksumAddress, schain_name: SchainName) -> int:
        return self.contract.functions.getBalance(address, schain_name).call()

    def check_user_balance(self, schain_name: SchainName, receiver: int) -> bool:
        return self.contract.functions.checkUserBalance(
            schain_name_to_hash(schain_name), receiver
        ).call()

    def get_recommended_recharge_amount(
        self,
        schain_name: SchainName,
        receiver: ChecksumAddress,
        transaction: TxParams | None = None,
    ) -> int:
        return self.contract.functions.getRecommendedRechargeAmount(
            schain_name_to_hash(schain_name), receiver
        ).call(transaction)

    def constant_setter_role(self) -> bool:
        return self.contract.functions.CONSTANT_SETTER_ROLE().call()

    def admin_role(self) -> bytes:
        return self.contract.functions.DEFAULT_ADMIN_ROLE().call()

    def has_role(self, role: bytes, address: ChecksumAddress) -> bool:
        return self.contract.functions.hasRole(role, address).call()

    @transaction_method
    def grant_role(self, role: bytes, address: ChecksumAddress) -> ContractFunction:
        return self.contract.functions.grantRole(role, address)

    def get_role_member(self, role: bytes, index: int) -> bytes:
        return self.contract.functions.getRoleMember(role, index).call()



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2021-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

import json
import os

from web3 import Web3

from skale.utils.web3_utils import get_eth_nonce, wait_for_receipt
from skale.wallets.common import BaseWallet

# Usage note: to change this contract update the code, compile it and put the new bytecode and
# new ABI below

FAKE_MULTISIG_CONTRACT = """
pragma solidity ^0.8.0;

contract FakeMultiSigWallet {

    receive() external payable {

    }

    function retrieve() external {
        payable(msg.sender).transfer(address(this).balance);
    }

    constructor() {

    }
}
"""

FAKE_MULTISIG_BYTECODE = '608060405234801561001057600080fd5b5060bc8061001f6000396000f3fe608060405260043610601f5760003560e01c80632e64cec114602a576025565b36602557005b600080fd5b348015603557600080fd5b50603c603e565b005b3373ffffffffffffffffffffffffffffffffffffffff166108fc479081150290604051600060405180830381858888f193505050501580156083573d6000803e3d6000fd5b5056fea26469706673582212205da5b248ec5ba69f49e730c2861325b8ad733be642688b5377ec25fc6da7e4fc64736f6c63430008070033'  # noqa

FAKE_MULTISIG_ABI = [
    {'inputs': [], 'stateMutability': 'nonpayable', 'type': 'constructor'},
    {
        'inputs': [],
        'name': 'retrieve',
        'outputs': [],
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    {'stateMutability': 'payable', 'type': 'receive'},
]

DIR_PATH = os.path.dirname(os.path.realpath(__file__))
FAKE_MULTISIG_DATA_DEFAULT_PATH = os.path.join(DIR_PATH, 'fake_multisig_data.json')
FAKE_MULTISIG_DATA_PATH = os.getenv('FAKE_MULTISIG_DATA_PATH') or FAKE_MULTISIG_DATA_DEFAULT_PATH

FAKE_MULTISIG_CONSTRUCTOR_GAS = 1000000


def deploy_fake_multisig_contract(web3: Web3, wallet: BaseWallet) -> None:
    print('Going to deploy simple payable contract')
    FakeMultisigContract = web3.eth.contract(abi=FAKE_MULTISIG_ABI, bytecode=FAKE_MULTISIG_BYTECODE)
    constructor = FakeMultisigContract.constructor()
    tx = constructor.build_transaction(
        {
            'nonce': get_eth_nonce(web3, wallet.address),
            'gasPrice': 3 * 10**9,
            'gas': FAKE_MULTISIG_CONSTRUCTOR_GAS,
        }
    )
    tx_hash = wallet.sign_and_send(tx)
    receipt = wait_for_receipt(web3, tx_hash)
    print(f'Sample contract successfully deployed: {receipt["contractAddress"]}')
    content = {'address': receipt['contractAddress'], 'abi': FAKE_MULTISIG_ABI}
    with open(FAKE_MULTISIG_DATA_PATH, 'w') as outfile:
        json.dump(content, outfile, indent=4)
    print(f'Sample contract data successfully saved to {FAKE_MULTISIG_DATA_PATH}')



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from dataclasses import dataclass

from skale.dataclasses.skaled_ports import SkaledPorts
from skale.types.node import NodeId, Port


@dataclass
class NodeInfo:
    """Dataclass that represents base info about the node"""

    node_id: NodeId
    name: str
    base_port: Port

    def calc_ports(self) -> dict[str, Port]:
        return {
            'httpRpcPort': Port(self.base_port + SkaledPorts.HTTP_JSON.value),
            'httpsRpcPort': Port(self.base_port + SkaledPorts.HTTPS_JSON.value),
            'wsRpcPort': Port(self.base_port + SkaledPorts.WS_JSON.value),
            'wssRpcPort': Port(self.base_port + SkaledPorts.WSS_JSON.value),
        }

    def to_dict(self) -> dict[str, NodeId | str | Port]:
        return {
            'nodeID': self.node_id,
            'nodeName': self.name,
            'basePort': self.base_port,
            **self.calc_ports(),
        }



"""SKALE token test"""

from tests.constants import TOKEN_TRANSFER_VALUE


def test_transfer(skale, empty_account):
    receiver_balance = skale.token.get_balance(empty_account.address)
    sender_balance = skale.token.get_balance(skale.wallet.address)

    assert receiver_balance == 0
    assert sender_balance != 0

    tx_res = skale.token.transfer(empty_account.address, TOKEN_TRANSFER_VALUE, wait_for=True)
    assert tx_res.receipt['status'] == 1

    receiver_balance_after = skale.token.get_balance(empty_account.address)
    sender_balance_after = skale.token.get_balance(skale.wallet.address)

    assert receiver_balance_after == TOKEN_TRANSFER_VALUE
    assert sender_balance_after == sender_balance - TOKEN_TRANSFER_VALUE


def test_transfer_wait_for(skale, empty_account):
    receiver_balance = skale.token.get_balance(empty_account.address)
    sender_balance = skale.token.get_balance(skale.wallet.address)

    assert receiver_balance == 0
    assert sender_balance != 0

    tx_res = skale.token.transfer(empty_account.address, TOKEN_TRANSFER_VALUE, wait_for=True)

    assert tx_res.receipt['status'] == 1

    receiver_balance_after = skale.token.get_balance(empty_account.address)
    sender_balance_after = skale.token.get_balance(skale.wallet.address)

    assert receiver_balance_after == TOKEN_TRANSFER_VALUE
    assert sender_balance_after == sender_balance - TOKEN_TRANSFER_VALUE


def test_get_balance(skale, empty_account):
    empty_balance = skale.token.get_balance(empty_account.address)
    not_empty_balance = skale.token.get_balance(skale.wallet.address)

    assert empty_balance == 0
    assert not_empty_balance != 0


def test_get_and_update_slashed_amount(skale):
    res = skale.token.get_and_update_slashed_amount(skale.wallet.address)
    assert res == 0



from skale.contracts.base_contract import transaction_method
from skale.contracts.ima.mainnet.linker import Linker
from skale.contracts.manager.contract_manager import ContractManager
from skale.contracts.skale_contract import SkaleContract

__all__ = ['Linker', 'ContractManager', 'SkaleContract', 'transaction_method']



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod
from typing import Optional

from eth_account.datastructures import SignedMessage, SignedTransaction
from eth_typing import ChecksumAddress, HexStr
from web3 import Web3
from web3.types import TxParams, TxReceipt, _Hash32

from skale.transactions.exceptions import ChainIdError
from skale.utils.web3_utils import DEFAULT_BLOCKS_TO_WAIT


def ensure_chain_id(tx_dict: TxParams, web3: Web3) -> None:
    if not tx_dict.get('chainId'):
        tx_dict['chainId'] = web3.eth.chain_id
    if not tx_dict.get('chainId'):
        raise ChainIdError('chainId must be in tx_dict (see EIP-155)')


class MessageNotSignedError(Exception):
    """
    Raised when signing message failed
    """

    pass


class BaseWallet(ABC):
    @abstractmethod
    def sign(self, tx_dict: TxParams) -> SignedTransaction:
        pass

    @abstractmethod
    def sign_and_send(
        self,
        tx_dict: TxParams,
        multiplier: Optional[float] = None,
        priority: Optional[int] = None,
        method: Optional[str] = None,
    ) -> HexStr:
        pass

    @abstractmethod
    def sign_hash(self, unsigned_hash: str) -> SignedMessage:
        pass

    @property
    @abstractmethod
    def address(self) -> ChecksumAddress:
        pass

    @property
    @abstractmethod
    def public_key(self) -> str:
        pass

    @abstractmethod
    def wait(self, tx: _Hash32, confirmation_blocks: int = DEFAULT_BLOCKS_TO_WAIT) -> TxReceipt:
        pass



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from skale_contracts.projects.skale_allocator import SkaleAllocatorContract
from skale_contracts.types import ContractName
from web3.constants import CHECKSUM_ADDRESSS_ZERO

from skale.contracts.skale_contract import SkaleContract


class AllocatorContract(SkaleContract):
    def init_contract(self, contract_name: ContractName) -> None:
        if contract_name == SkaleAllocatorContract.ESCROW:
            self.address = CHECKSUM_ADDRESSS_ZERO
        else:
            self.address = self.skale.instance.get_contract_address(contract_name)
        self.contract = self.skale.web3.eth.contract(
            address=self.address,
            abi=self.skale.instance.abi[contract_name],
        )



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2025-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

import hashlib
import json
import logging
from collections.abc import Mapping as ABCMapping
from dataclasses import dataclass, field
from typing import Any, Mapping

from redis import Redis
from redis.exceptions import RedisError
from web3.middleware.base import Web3Middleware

logger = logging.getLogger(__name__)


RPC_CACHE_METHODS: tuple[str, ...] = (
    'eth_call',
    'eth_getCode',
    'eth_getStorageAt',
    'eth_chainId',
    'eth_getBlockByNumber',
    'eth_gasPrice',
    'web3_clientVersion',
)

METHOD_TTL_POLICY: dict[str, int] = {
    'eth_call': 5,
    'eth_getCode': 30,
    'eth_getStorageAt': 30,
    'eth_chainId': 600,
    'eth_getBlockByNumber': 60,
    'eth_gasPrice': 5,
    'web3_clientVersion': 600,
}


@dataclass(frozen=True, slots=True)
class RedisCacheConfig:
    redis_url: str
    cache_methods: frozenset[str] = field(default_factory=lambda: frozenset(RPC_CACHE_METHODS))
    bypass_addresses: frozenset[str] = frozenset()
    default_ttl_seconds: int = 1
    key_prefix: str = 'rpc-cache:v1'
    method_ttl_policy: Mapping[str, int] = field(default_factory=lambda: METHOD_TTL_POLICY)


def ttl_policy(
    method: str,
    params: list[Any],
    default_ttl_seconds: int,
    method_ttl_policy: Mapping[str, int] | None = None,
) -> int:
    method_default_ttl = (
        method_ttl_policy[method]
        if method_ttl_policy is not None and method in method_ttl_policy
        else default_ttl_seconds
    )

    if method_default_ttl <= 0:
        return 0

    block_id = _extract_block_identifier(method, params)
    if isinstance(block_id, int):
        return 60
    if isinstance(block_id, str):
        if block_id.startswith('0x'):
            return 60
        if block_id in {'latest', 'safe', 'finalized'}:
            return method_default_ttl
        if block_id == 'pending':
            return 0

    return method_default_ttl


def redis_cache_middleware(config: RedisCacheConfig) -> type[Web3Middleware]:
    redis = Redis.from_url(config.redis_url)
    cache_methods = {m for m in config.cache_methods}
    bypass_addresses = {_normalize_address(a) for a in config.bypass_addresses}
    disallowed = {
        'eth_getLogs',
        'eth_sendRawTransaction',
        'eth_sendTransaction',
        'personal_sendTransaction',
        'eth_sign',
        'eth_signTransaction',
        'eth_signTypedData',
        'eth_signTypedData_v3',
        'eth_signTypedData_v4',
    }

    class _RedisCacheMiddleware(Web3Middleware):
        def wrap_make_request(self, make_request):
            endpoint = getattr(self._w3.provider, 'endpoint_uri', '') or ''
            endpoint_hash = hashlib.sha256(endpoint.encode('utf-8')).hexdigest()[:16]

            def request(method: str, params: list[Any]):
                if method in disallowed or method not in cache_methods:
                    return make_request(method, params)

                if method == 'eth_getTransactionCount' and _is_pending_tx_count(params):
                    return make_request(method, params)

                if method in {'eth_call', 'eth_estimateGas'} and _is_bypassed_address(
                    params, bypass_addresses
                ):
                    return make_request(method, params)

                ttl_seconds = ttl_policy(
                    method,
                    params,
                    config.default_ttl_seconds,
                    config.method_ttl_policy,
                )
                if ttl_seconds <= 0:
                    return make_request(method, params)

                key = _cache_key(
                    prefix=config.key_prefix,
                    endpoint_hash=endpoint_hash,
                    method=method,
                    params=params,
                )

                try:
                    cached = redis.get(key)
                except (OSError, RedisError):
                    cached = None

                if cached:
                    try:
                        return json.loads(cached)
                    except (TypeError, UnicodeDecodeError, json.JSONDecodeError):
                        pass

                response = make_request(method, params)
                if isinstance(response, dict) and 'error' not in response and 'result' in response:
                    try:
                        payload = json.dumps(
                            response,
                            separators=(',', ':'),
                            sort_keys=True,
                            default=_json_default,
                        ).encode('utf-8')
                        redis.setex(key, ttl_seconds, payload)
                    except (RedisError, ValueError, UnicodeEncodeError):
                        pass

                return response

            return request

    return _RedisCacheMiddleware


def _cache_key(prefix: str, endpoint_hash: str, method: str, params: list[Any]) -> str:
    body = json.dumps(
        {'method': method, 'params': params},
        separators=(',', ':'),
        sort_keys=True,
        default=_json_default,
    ).encode('utf-8')
    request_hash = hashlib.sha256(body).hexdigest()
    return f'{prefix}:{endpoint_hash}:{request_hash}'


def _json_default(value: Any) -> Any:
    if isinstance(value, ABCMapping):
        return dict(value)
    if isinstance(value, (bytes, bytearray, memoryview)):
        return '0x' + bytes(value).hex()
    if hasattr(value, 'hex'):
        try:
            return value.hex()
        except (TypeError, ValueError):
            return str(value)
    return str(value)


def _normalize_address(address: str) -> str:
    address = address.strip().lower()
    if address.startswith('0x'):
        return address
    return '0x' + address


def _is_bypassed_address(params: list[Any], bypass_addresses: set[str]) -> bool:
    if not bypass_addresses or not params:
        return False
    tx = params[0]
    if not isinstance(tx, dict):
        return False
    to_addr = tx.get('to')
    if not isinstance(to_addr, str):
        return False
    return _normalize_address(to_addr) in bypass_addresses


def _is_pending_tx_count(params: list[Any]) -> bool:
    if len(params) < 2:
        return False
    block_id = params[1]
    return block_id == 'pending'


def _extract_block_identifier(method: str, params: list[Any]) -> Any:
    if method in {'eth_call', 'eth_estimateGas'}:
        return params[1] if len(params) > 1 else 'latest'
    if method in {'eth_getBalance', 'eth_getCode', 'eth_getStorageAt', 'eth_getTransactionCount'}:
        return params[1] if len(params) > 1 else 'latest'
    return None



"""SKALE account tools test"""

from unittest import mock

from skale.utils.account_tools import (
    check_ether_balance,
    check_skale_balance,
    generate_account,
    generate_accounts,
    send_eth,
    send_tokens,
)
from skale.utils.constants import GAS_PRICE_COEFFICIENT
from skale.utils.web3_utils import get_eth_nonce
from skale.wallets.web3_wallet import Web3Wallet, generate_wallet
from tests.constants import ETH_TRANSFER_VALUE, N_TEST_WALLETS, TOKEN_TRANSFER_VALUE


def test_send_tokens(skale, empty_account):
    sender_balance = skale.token.get_balance(skale.wallet.address)

    send_tokens(skale, empty_account.address, TOKEN_TRANSFER_VALUE)

    receiver_balance_after = skale.token.get_balance(empty_account.address)
    sender_balance_after = skale.token.get_balance(skale.wallet.address)

    token_transfer_value_wei = skale.web3.to_wei(TOKEN_TRANSFER_VALUE, 'ether')

    assert receiver_balance_after == token_transfer_value_wei
    assert sender_balance_after == sender_balance - token_transfer_value_wei


def test_send_eth(skale, empty_account):
    sender_balance = check_ether_balance(skale.web3, skale.wallet.address)

    send_eth(skale.web3, skale.wallet, empty_account.address, ETH_TRANSFER_VALUE)

    receiver_balance_after = check_ether_balance(skale.web3, empty_account.address)
    sender_balance_after = check_ether_balance(skale.web3, skale.wallet.address)

    assert receiver_balance_after == ETH_TRANSFER_VALUE
    # check that sender_balance_after
    # have decreased by ETH_TRANSFER_VALUE and some gas
    assert (
        sender_balance - 2 * ETH_TRANSFER_VALUE
        < sender_balance_after
        < sender_balance - ETH_TRANSFER_VALUE
    )


def test_send_eth_with_gas_price(skale, empty_account):
    def get_signed_tx_with_custom_gas_price(gas_price):
        wei_amount = skale.web3.to_wei(ETH_TRANSFER_VALUE, 'ether')
        return skale.wallet.sign(
            {
                'to': empty_account.address,
                'value': wei_amount,
                'gasPrice': gas_price,
                'gas': 22000,
                'nonce': get_eth_nonce(skale.web3, skale.wallet.address),
            }
        )

    custom_default_gas_price = 101 * 10**9
    with mock.patch(
        'skale.utils.account_tools.default_gas_price', return_value=custom_default_gas_price
    ):
        receipt = send_eth(
            skale.web3, skale.wallet, empty_account.address, ETH_TRANSFER_VALUE, wait_for=True
        )
        assert receipt['effectiveGasPrice'] == custom_default_gas_price

    # Send ether with default avg gas price
    avg_gas_price = skale.web3.eth.gas_price * GAS_PRICE_COEFFICIENT
    receipt = send_eth(
        skale.web3, skale.wallet, empty_account.address, ETH_TRANSFER_VALUE, wait_for=True
    )
    assert receipt['effectiveGasPrice'] == avg_gas_price


def test_generate_account(skale):
    account = generate_account(skale.web3)
    assert account.get('address') is not None
    assert account.get('private_key') is not None


def test_generate_accounts(skale):
    results = generate_accounts(
        skale, skale.wallet, N_TEST_WALLETS, TOKEN_TRANSFER_VALUE, ETH_TRANSFER_VALUE, debug=True
    )
    assert len(results) == N_TEST_WALLETS

    test_address = results[-1]['address']
    eth_balance = check_ether_balance(skale.web3, test_address)
    token_balance_test = check_skale_balance(skale, test_address)
    token_balance = skale.token.get_balance(test_address)

    token_transfer_value_wei = skale.web3.to_wei(TOKEN_TRANSFER_VALUE, 'ether')

    assert TOKEN_TRANSFER_VALUE == token_balance_test
    assert eth_balance == ETH_TRANSFER_VALUE
    assert token_balance == token_transfer_value_wei


def test_generate_wallet(skale):
    wallet = generate_wallet(skale.web3)
    assert isinstance(wallet, Web3Wallet)
    assert wallet.address
    assert wallet.address != skale.wallet.address



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from eth_typing import ChecksumAddress
from web3.contract.contract import ContractFunction

from skale.contracts.base_contract import transaction_method
from skale.contracts.skale_manager_contract import SkaleManagerContract


class DelegationPeriodManager(SkaleManagerContract):
    """Wrapper for DelegationPeriodManager.sol functions"""

    @transaction_method
    def set_delegation_period(self, months_count: int, stake_multiplier: int) -> 'ContractFunction':
        return self.contract.functions.setDelegationPeriod(
            monthsCount=months_count, stakeMultiplier=stake_multiplier
        )

    def is_delegation_period_allowed(self, months_count: int) -> bool:
        return bool(
            self.contract.functions.isDelegationPeriodAllowed(monthsCount=months_count).call()
        )

    @transaction_method
    def grant_role(self, role: bytes, address: ChecksumAddress) -> 'ContractFunction':
        return self.contract.functions.grantRole(role, address)

    def delegation_period_setter_role(self) -> bytes:
        return bytes(self.contract.functions.DELEGATION_PERIOD_SETTER_ROLE().call())

    def has_role(self, role: bytes, address: ChecksumAddress) -> bool:
        return bool(self.contract.functions.hasRole(role, address).call())



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2025-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

import pytest


@pytest.mark.parametrize('number_of_nodes', [1])
def test_create_exit_request_and_fetch(fair, fair_active_nodes):
    node = fair.nodes.get_by_address(fair_active_nodes[0].address)
    node_id = node.id
    fair.status.whitelist_node(node_id)
    fair.staking.stake(node_id, value=10_000)

    initial_count = fair.staking.get_my_exit_requests_count()

    fair.staking.request_retrieve(node_id, 1_000)

    new_count = fair.staking.get_my_exit_requests_count()
    assert new_count == initial_count + 1

    my_address = fair.wallet.address
    requests = fair.staking.get_exit_requests_for(my_address)
    assert len(requests) == new_count
    last_request = requests[-1]
    assert last_request.user == my_address
    assert last_request.node_id == node_id
    assert last_request.amount == 1_000
    assert isinstance(last_request.unlock_date, int)


@pytest.mark.parametrize('number_of_nodes', [1])
def test_exit_request_unlock_check(fair, fair_active_nodes):
    node = fair.nodes.get_by_address(fair_active_nodes[0].address)
    node_id = node.id
    fair.status.whitelist_node(node_id)
    fair.staking.stake(node_id, value=10_000)

    fair.staking.request_retrieve(node_id, 500)

    my_address = fair.wallet.address
    request = fair.staking.get_exit_request_at(my_address, 0)
    unlocked_flag = fair.staking.is_request_unlocked(request.request_id)
    assert isinstance(unlocked_flag, bool)


@pytest.mark.parametrize('number_of_nodes', [1])
def test_retrieve_all_request(fair, fair_active_nodes):
    node = fair.nodes.get_by_address(fair_active_nodes[0].address)
    node_id = node.id
    fair.status.whitelist_node(node_id)
    fair.staking.stake(node_id, value=5_000)

    fair.staking.request_retrieve_all(node_id)
    my_address = fair.wallet.address
    count = fair.staking.get_my_exit_requests_count()
    assert count >= 1

    request = fair.staking.get_exit_request_at(my_address, count - 1)
    assert request.amount > 0
